@@ -24,7 +24,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: main.c,v 1.71 2008/04/03 20:51:45 sobomax Exp $
+ * $Id: main.c,v 1.87 2008/12/24 10:46:03 sobomax Exp $
  *
  */
 
@@ -42,9 +42,11 @@
 #include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <grp.h>
 #include <limits.h>
 #include <netdb.h>
 #include <poll.h>
+#include <pwd.h>
 #include <sched.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -93,9 +95,10 @@ static void
 usage(void)
 {
 
-    fprintf(stderr, "usage: rtpproxy [-2fvF] [-l addr1[/addr2]] "
+    fprintf(stderr, "usage: rtpproxy [-2fvFiPa] [-l addr1[/addr2]] "
       "[-6 addr1[/addr2]] [-s path]\n\t[-t tos] [-r rdir [-S sdir]] [-T ttl] "
-      "[-L nfiles] [-m port_min]\n\t[-M port_max] [-u uname[:gname]]\n");
+      "[-L nfiles] [-m port_min]\n\t[-M port_max] [-u uname[:gname]] "
+      "[-n timeout_socket] [-d log_level]\n");
     exit(1);
 }
 
@@ -122,6 +125,8 @@ init_config(struct cfg *cf, int argc, char **argv)
 {
     int ch, i;
     char *bh[2], *bh6[2], *cp;
+    struct passwd *pp;
+    struct group *gp;
 
     bh[0] = bh[1] = bh6[0] = bh6[1] = NULL;
 
@@ -131,11 +136,17 @@ init_config(struct cfg *cf, int argc, char **argv)
     cf->max_ttl = SESSION_TIMEOUT;
     cf->tos = TOS;
     cf->rrtcp = 1;
+    cf->ttl_mode = TTL_UNIFIED;
+    cf->log_level = LOG_LEVEL;
+
+    cf->timeout_handler.socket_name = NULL;
+    cf->timeout_handler.fd = -1;
+    cf->timeout_handler.connected = 0;
 
     if (getrlimit(RLIMIT_NOFILE, &(cf->nofile_limit)) != 0)
 	err(1, "getrlimit");
 
-    while ((ch = getopt(argc, argv, "vf2Rl:6:s:S:t:r:p:T:L:m:M:u:F")) != -1)
+    while ((ch = getopt(argc, argv, "vf2Rl:6:s:S:t:r:p:T:L:m:M:u:Fin:Pad:")) != -1)
 	switch (ch) {
 	case 'f':
 	    cf->nodaemon = 1;
@@ -242,12 +253,57 @@ init_config(struct cfg *cf, int argc, char **argv)
 		cp++;
 	    }
 	    cf->run_gname = cp;
+	    cf->run_uid = -1;
+	    cf->run_gid = -1;
+	    if (cf->run_uname != NULL) {
+		pp = getpwnam(cf->run_uname);
+		if (pp == NULL)
+		    err(1, "can't find ID for the user: %s", cf->run_uname);
+		cf->run_uid = pp->pw_uid;
+		if (cf->run_gname == NULL)
+		    cf->run_gid = pp->pw_gid;
+	    }
+	    if (cf->run_gname != NULL) {
+		gp = getgrnam(cf->run_gname);
+		if (gp == NULL)
+		    err(1, "can't find ID for the group: %s", cf->run_gname);
+		cf->run_gid = gp->gr_gid;
+	    }
 	    break;
 
 	case 'F':
 	    cf->no_check = 1;
 	    break;
 
+	case 'i':
+	    cf->ttl_mode = TTL_INDEPENDENT;
+	    break;
+
+	case 'n':
+	    if(strncmp("unix:", optarg, 5) == 0)
+		optarg += 5;
+	    if(strlen(optarg) == 0)
+		errx(1, "timeout notification socket name too short");
+	    cf->timeout_handler.socket_name = (char *)malloc(strlen(optarg) + 1);
+	    if(cf->timeout_handler.socket_name == NULL)
+		err(1, "can't allocate memory");
+	    strcpy(cf->timeout_handler.socket_name, optarg);
+	    break;
+
+	case 'P':
+	    cf->record_pcap = 1;
+	    break;
+
+	case 'a':
+	    cf->record_all = 1;
+	    break;
+
+	case 'd':
+	    cf->log_level = rtpp_log_str2lvl(optarg);
+	    if (cf->log_level == -1)
+		errx(1, "%s: invalid log level", optarg);
+	    break;
+
 	case '?':
 	default:
 	    usage();
@@ -269,6 +325,19 @@ init_config(struct cfg *cf, int argc, char **argv)
 	}
     }
 
+    /* make sure that port_min and port_max are even */
+    if ((cf->port_min % 2) != 0)
+	cf->port_min++;
+    if ((cf->port_max % 2) != 0) {
+	cf->port_max--;
+    } else {
+	/*
+	 * If port_max is already even then there is no
+	 * "room" for the RTCP port, go back by two ports.
+	 */
+	cf->port_max -= 2;
+    }
+
     if (cf->port_min <= 0 || cf->port_min > 65535)
 	errx(1, "invalid value of the port_min argument, "
 	  "not in the range 1-65535");
@@ -278,13 +347,6 @@ init_config(struct cfg *cf, int argc, char **argv)
     if (cf->port_min > cf->port_max)
 	errx(1, "port_min should be less than port_max");
 
-    /* make sure that port_min and port_max are even */
-    if ((cf->port_min % 2) != 0)
-	cf->port_min++;
-    if ((cf->port_max % 2) != 0)
-	cf->port_max--;
-
-    cf->nextport[0] = cf->nextport[1] = cf->port_min;
     cf->sessions = malloc((sizeof cf->sessions[0]) *
       (((cf->port_max - cf->port_min + 1) * 2) + 1));
     cf->rtp_servers =  malloc((sizeof cf->rtp_servers[0]) *
@@ -352,7 +414,7 @@ init_controlfd(struct cfg *cf)
     if (cf->umode == 0) {
 	unlink(cmd_sock);
 	memset(&ifsun, '\0', sizeof ifsun);
-#if !defined(__linux__) && !defined(__solaris__)
+#if defined(HAVE_SOCKADDR_SUN_LEN)
 	ifsun.sun_len = strlen(cmd_sock);
 #endif
 	ifsun.sun_family = AF_LOCAL;
@@ -364,6 +426,9 @@ init_controlfd(struct cfg *cf)
 	  sizeof controlfd);
 	if (bind(controlfd, sstosa(&ifsun), sizeof ifsun) < 0)
 	    err(1, "can't bind to a socket");
+	if ((cf->run_uname != NULL || cf->run_gname != NULL) &&
+	  chown(cmd_sock, cf->run_uid, cf->run_gid) == -1)
+	    err(1, "can't set owner of the socket");
 	if (listen(controlfd, 32) != 0)
 	    err(1, "can't listen on a socket");
     } else {
@@ -389,7 +454,7 @@ init_controlfd(struct cfg *cf)
 }
 
 static void
-process_rtp_servers(struct cfg *cf, double ctime)
+process_rtp_servers(struct cfg *cf, double dtime)
 {
     int j, k, sidx, len, skipfd;
     struct rtpp_session *sp;
@@ -408,7 +473,7 @@ process_rtp_servers(struct cfg *cf, double ctime)
 	for (sidx = 0; sidx < 2; sidx++) {
 	    if (sp->rtps[sidx] == NULL || sp->addr[sidx] == NULL)
 		continue;
-	    while ((len = rtp_server_get(sp->rtps[sidx], ctime)) != RTPS_LATER) {
+	    while ((len = rtp_server_get(sp->rtps[sidx], dtime)) != RTPS_LATER) {
 		if (len == RTPS_EOF) {
 		    rtp_server_free(sp->rtps[sidx]);
 		    sp->rtps[sidx] = NULL;
@@ -431,7 +496,7 @@ process_rtp_servers(struct cfg *cf, double ctime)
 
 static void
 rxmit_packets(struct cfg *cf, struct rtpp_session *sp, int ridx,
-  double ctime)
+  double dtime)
 {
     int ndrain, i, port;
     struct rtp_packet *packet = NULL;
@@ -444,7 +509,9 @@ rxmit_packets(struct cfg *cf, struct rtpp_session *sp, int ridx,
 	packet = rtp_recv(sp->fds[ridx]);
 	if (packet == NULL)
 	    break;
-	packet->rtime = ctime;
+	packet->laddr = sp->laddr[ridx];
+	packet->rport = sp->ports[ridx];
+	packet->rtime = dtime;
 
 	i = 0;
 	if (sp->addr[ridx] != NULL) {
@@ -458,8 +525,12 @@ rxmit_packets(struct cfg *cf, struct rtpp_session *sp, int ridx,
 			 */
 			continue;
 		    }
-		    /* Signal that an address have to be updated */
+		    /* Signal that an address has to be updated */
 		    i = 1;
+		} else if (sp->canupdate[ridx] != 0 &&
+		  sp->last_update[ridx] != 0 &&
+		  dtime - sp->last_update[ridx] > UPDATE_WINDOW) {
+		    sp->canupdate[ridx] = 0;
 		}
 	    } else {
 		/*
@@ -499,7 +570,10 @@ rxmit_packets(struct cfg *cf, struct rtpp_session *sp, int ridx,
 	if (i != 0) {
 	    sp->untrusted_addr[ridx] = 1;
 	    memcpy(sp->addr[ridx], &packet->raddr, packet->rlen);
-	    sp->canupdate[ridx] = 0;
+	    if (sp->prev_addr[ridx] == NULL || memcmp(sp->prev_addr[ridx],
+	      &packet->raddr, packet->rlen) != 0) {
+	        sp->canupdate[ridx] = 0;
+	    }
 
 	    port = ntohs(satosin(&packet->raddr)->sin_port);
 
@@ -556,7 +630,7 @@ send_packet(struct cfg *cf, struct rtpp_session *sp, int ridx,
 {
     int i, sidx;
 
-    GET_RTP(sp)->ttl = cf->max_ttl;
+    GET_RTP(sp)->ttl[ridx] = cf->max_ttl;
 
     /* Select socket for sending packet out. */
     sidx = (ridx == 0) ? 1 : 0;
@@ -570,7 +644,7 @@ send_packet(struct cfg *cf, struct rtpp_session *sp, int ridx,
     } else {
 	sp->pcount[2]++;
 	for (i = (cf->dmode && packet->size < LBR_THRS) ? 2 : 1; i > 0; i--) {
-	    sendto(sp->fds[sidx], packet->buf, packet->size, 0, sp->addr[sidx],
+	    sendto(sp->fds[sidx], packet->data.buf, packet->size, 0, sp->addr[sidx],
 	      SA_LEN(sp->addr[sidx]));
 	}
     }
@@ -580,7 +654,7 @@ send_packet(struct cfg *cf, struct rtpp_session *sp, int ridx,
 }
 
 static void
-process_rtp(struct cfg *cf, double ctime, int alarm_tick)
+process_rtp(struct cfg *cf, double dtime, int alarm_tick)
 {
     int readyfd, skipfd, ridx;
     struct rtpp_session *sp;
@@ -593,11 +667,15 @@ process_rtp(struct cfg *cf, double ctime, int alarm_tick)
 
 	if (alarm_tick != 0 && sp != NULL && sp->rtcp != NULL &&
 	  sp->sidx[0] == readyfd) {
-	    if (sp->ttl == 0) {
+	    if (get_ttl(sp) == 0) {
 		rtpp_log_write(RTPP_LOG_INFO, sp->log, "session timeout");
+		do_timeout_notification(sp, 1);
 		remove_session(cf, sp);
 	    } else {
-		sp->ttl--;
+		if (sp->ttl[0] != 0)
+		    sp->ttl[0]--;
+		if (sp->ttl[1] != 0)
+		    sp->ttl[1]--;
 	    }
 	}
 
@@ -625,9 +703,9 @@ process_rtp(struct cfg *cf, double ctime, int alarm_tick)
 
 	if (sp->complete != 0) {
 	    if ((cf->pfds[readyfd].revents & POLLIN) != 0)
-		rxmit_packets(cf, sp, ridx, ctime);
+		rxmit_packets(cf, sp, ridx, dtime);
 	    if (sp->resizers[ridx].output_nsamples > 0) {
-		while ((packet = rtp_resizer_get(&sp->resizers[ridx], ctime)) != NULL) {
+		while ((packet = rtp_resizer_get(&sp->resizers[ridx], dtime)) != NULL) {
 		    send_packet(cf, sp, ridx, packet);
 		    rtp_packet_free(packet);
 		}
@@ -639,7 +717,7 @@ process_rtp(struct cfg *cf, double ctime, int alarm_tick)
 }
 
 static void
-process_commands(struct cfg *cf)
+process_commands(struct cfg *cf, double dtime)
 {
     int controlfd, i;
     socklen_t rlen;
@@ -661,7 +739,7 @@ process_commands(struct cfg *cf)
 	} else {
 	    controlfd = cf->pfds[0].fd;
 	}
-	i = handle_command(cf, controlfd);
+	i = handle_command(cf, controlfd, dtime);
 	if (cf->umode == 0) {
 	    close(controlfd);
 	}
@@ -679,22 +757,23 @@ main(int argc, char **argv)
 
     memset(&cf, 0, sizeof(cf));
 
-    init_hash_table(&cf);
-
     init_config(&cf, argc, argv);
 
+    seedrandom();
+
+    init_hash_table(&cf);
+    init_port_table(&cf);
+
     controlfd = init_controlfd(&cf);
 
-#if !defined(__solaris__)
     if (cf.nodaemon == 0) {
-	if (daemon(0, 0) == -1)
+	if (rtpp_daemon(0, 0) == -1)
 	    err(1, "can't switch into daemon mode");
 	    /* NOTREACHED */
     }
-#endif
 
+    glog = cf.glog = rtpp_log_open(&cf, "rtpproxy", NULL, LF_REOPEN);
     atexit(ehandler);
-    glog = cf.glog = rtpp_log_open("rtpproxy", NULL, LF_REOPEN);
     rtpp_log_write(RTPP_LOG_INFO, cf.glog, "rtpproxy started, pid %d", getpid());
 
     i = open(pid_file, O_WRONLY | O_CREAT | O_TRUNC, DEFFILEMODE);
@@ -719,7 +798,7 @@ main(int argc, char **argv)
     signal(SIGUSR2, fatsignal);
 
     if (cf.run_uname != NULL || cf.run_gname != NULL) {
-	if (drop_privileges(&cf, cf.run_uname, cf.run_gname) != 0) {
+	if (drop_privileges(&cf) != 0) {
 	    rtpp_log_ewrite(RTPP_LOG_ERR, cf.glog,
 	      "can't switch to requested user/group");
 	    exit(1);
@@ -741,18 +820,18 @@ main(int argc, char **argv)
 	    timeout = RTPS_TICKS_MIN;
 	else
 	    timeout = TIMETICK * 1000;
-	eptime = getctime();
+	eptime = getdtime();
 	delay = (eptime - sptime) * 1000000.0;
 	if (delay < (1000000 / POLL_LIMIT)) {
 	    usleep((1000000 / POLL_LIMIT) - delay);
-	    sptime = getctime();
+	    sptime = getdtime();
 	} else {
 	    sptime = eptime;
 	}
 	i = poll(cf.pfds, cf.nsessions, timeout);
 	if (i < 0 && errno == EINTR)
 	    continue;
-	eptime = getctime();
+	eptime = getdtime();
 	if (cf.rtp_nsessions > 0) {
 	    process_rtp_servers(&cf, eptime);
 	}
@@ -764,7 +843,7 @@ main(int argc, char **argv)
 	}
 	process_rtp(&cf, eptime, alarm_tick);
 	if (i > 0) {
-	    process_commands(&cf);
+	    process_commands(&cf, eptime);
 	}
     }
 
