@@ -24,7 +24,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: rtpp_command.c,v 1.6 2008/04/03 23:05:41 sobomax Exp $
+ * $Id: rtpp_command.c,v 1.24 2009/01/12 10:56:59 sobomax Exp $
  *
  */
 
@@ -33,6 +33,7 @@
 #include <sys/uio.h>
 #include <netinet/in.h>
 #include <assert.h>
+#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
@@ -60,15 +61,18 @@ struct proto_cap proto_caps[] = {
     { "20071116", "Support for RTP re-packetization" },
     { "20071218", "Support for forking (copying) RTP stream" },
     { "20080403", "Support for RTP statistics querying" },
+    { "20081102", "Support for setting codecs in the update/lookup command" },
+    { "20081224", "Support for session timeout notifications" },
     { NULL, NULL }
 };
 
 static int create_twinlistener(struct cfg *, struct sockaddr *, int, int *);
-static int create_listener(struct cfg *, struct sockaddr *, int, int *, int *);
+static int create_listener(struct cfg *, struct sockaddr *, int *, int *);
 static int handle_delete(struct cfg *, char *, char *, char *, int);
 static void handle_noplay(struct cfg *, struct rtpp_session *, int);
 static int handle_play(struct cfg *, struct rtpp_session *, int, char *, char *, int);
-static void handle_record(struct cfg *, struct rtpp_session *, int, char *);
+static void handle_copy(struct cfg *, struct rtpp_session *, int, char *);
+static int handle_record(struct cfg *, char *, char *, char *);
 static void handle_query(struct cfg *, int, struct sockaddr_storage *,
   socklen_t, char *, struct rtpp_session *, int);
 
@@ -118,28 +122,23 @@ failure:
 }
 
 static int
-create_listener(struct cfg *cf, struct sockaddr *ia,  int startport,
-  int *port, int *fds)
+create_listener(struct cfg *cf, struct sockaddr *ia, int *port, int *fds)
 {
-    int i, init, rval;
+    int i, idx, rval;
 
     for (i = 0; i < 2; i++)
 	fds[i] = -1;
 
-    init = 0;
-    if (startport < cf->port_min || startport > cf->port_max)
-	startport = cf->port_min;
-    for (*port = startport; *port != startport || init == 0; (*port) += 2) {
-	init = 1;
+    for (i = 1; i < cf->port_table_len; i++) {
+	idx = (cf->port_table_idx + i) % cf->port_table_len;
+	*port = cf->port_table[idx];
 	rval = create_twinlistener(cf, ia, *port, fds);
-	if (rval != 0) {
-	    if (rval == -1)
-		break;
-	    if (*port >= cf->port_max)
-		*port = cf->port_min - 2;
-	    continue;
+	if (rval == 0) {
+	    cf->port_table_idx = idx;
+	    return 0;
 	}
-	return 0;
+	if (rval == -1)
+	    break;
     }
     return -1;
 }
@@ -242,7 +241,7 @@ handle_nomem(struct cfg *cf, int fd, struct sockaddr_storage *raddr,
 }
 
 int
-handle_command(struct cfg *cf, int controlfd)
+handle_command(struct cfg *cf, int controlfd, double dtime)
 {
     int len, argc, i, j, pidx, asymmetric;
     int external, pf, lidx, playcount, weak;
@@ -250,7 +249,7 @@ handle_command(struct cfg *cf, int controlfd)
     socklen_t rlen;
     char buf[1024 * 8];
     char *cp, *call_id, *from_tag, *to_tag, *addr, *port, *cookie;
-    char *pname, *codecs, *recording_name;
+    char *pname, *codecs, *recording_name, *t;
     struct rtpp_session *spa, *spb;
     char **ap, *argv[10];
     const char *rname;
@@ -258,6 +257,8 @@ handle_command(struct cfg *cf, int controlfd)
     struct sockaddr_storage raddr;
     int requested_nsamples;
     enum {DELETE, RECORD, PLAY, NOPLAY, COPY, UPDATE, LOOKUP, QUERY} op;
+    int max_argc;
+    char *socket_name_u, *notify_tag;
 
     requested_nsamples = -1;
     ia[0] = ia[1] = NULL;
@@ -266,6 +267,7 @@ handle_command(struct cfg *cf, int controlfd)
     lidx = 1;
     fds[0] = fds[1] = -1;
     recording_name = NULL;
+    socket_name_u = notify_tag = NULL;
 
     if (cf->umode == 0) {
 	for (;;) {
@@ -291,7 +293,7 @@ handle_command(struct cfg *cf, int controlfd)
     cp = buf;
     argc = 0;
     memset(argv, 0, sizeof(argv));
-    for (ap = argv; (*ap = strsep(&cp, "\r\n\t ")) != NULL;)
+    for (ap = argv; (*ap = rtpp_strsep(&cp, "\r\n\t ")) != NULL;)
 	if (**ap != '\0') {
 	    argc++;
 	    if (++ap >= &argv[10])
@@ -319,6 +321,7 @@ handle_command(struct cfg *cf, int controlfd)
     switch (argv[0][0]) {
     case 'u':
     case 'U':
+	/* U[opts] callid remote_ip remote_port from_tag [to_tag] */
 	op = UPDATE;
 	rname = "update/create";
 	break;
@@ -337,7 +340,13 @@ handle_command(struct cfg *cf, int controlfd)
 
     case 'p':
     case 'P':
-	/* P callid pname codecs from_tag to_tag */
+	/*
+	 * P callid pname codecs from_tag to_tag
+	 *
+	 *   <codecs> could be either comma-separated list of supported
+	 *   payload types or word "session" (without quotes), in which
+	 *   case list saved on last session update will be used instead.
+	 */
 	op = PLAY;
 	rname = "play";
 	playcount = 1;
@@ -438,10 +447,10 @@ handle_command(struct cfg *cf, int controlfd)
 
 	    len += sprintf(buf + len,
 	      "%s/%s: caller = %s:%d/%s, callee = %s:%d/%s, "
-	      "stats = %lu/%lu/%lu/%lu, ttl = %d\n",
+	      "stats = %lu/%lu/%lu/%lu, ttl = %d/%d\n",
 	      spb->call_id, spb->tag, addrs[0], spb->ports[1], addrs[1],
 	      addrs[2], spb->ports[0], addrs[3], spa->pcount[0], spa->pcount[1],
-	      spa->pcount[2], spa->pcount[3], spb->ttl);
+	      spa->pcount[2], spa->pcount[3], spb->ttl[0], spb->ttl[1]);
 	    if (len + 512 > sizeof(buf)) {
 		doreply(cf, controlfd, buf, len, &raddr, rlen);
 		len = 0;
@@ -458,6 +467,23 @@ handle_command(struct cfg *cf, int controlfd)
 	rname = "query";
 	break;
 
+    case 'x':
+    case 'X':
+        /* Delete all active sessions */
+        rtpp_log_write(RTPP_LOG_INFO, cf->glog, "deleting all active sessions");
+        for (i = 1; i < cf->nsessions; i++) {
+	    spa = cf->sessions[i];
+	    if (spa == NULL || spa->sidx[0] != i)
+		continue;
+	    /* Skip RTCP twin session */
+	    if (spa->rtcp != NULL) {
+		remove_session(cf, spa);
+	    }
+        }
+        reply_ok(cf, controlfd, &raddr, rlen, cookie);
+        return 0;
+        break;
+
     default:
 	rtpp_log_write(RTPP_LOG_ERR, cf->glog, "unknown command");
 	reply_error(cf, controlfd, &raddr, rlen, cookie, 3);
@@ -465,7 +491,8 @@ handle_command(struct cfg *cf, int controlfd)
     }
     call_id = argv[1];
     if (op == UPDATE || op == LOOKUP || op == PLAY) {
-	if (argc < 5 || argc > 6) {
+	max_argc = (op == UPDATE ? 8 : 6);
+	if (argc < 5 || argc > max_argc) {
 	    rtpp_log_write(RTPP_LOG_ERR, cf->glog, "command syntax error");
 	    reply_error(cf, controlfd, &raddr, rlen, cookie, 4);
 	    return 0;
@@ -474,6 +501,22 @@ handle_command(struct cfg *cf, int controlfd)
 	to_tag = argv[5];
 	if (op == PLAY && argv[0][1] != '\0')
 	    playcount = atoi(argv[0] + 1);
+	if (op == UPDATE && argc > 6) {
+	    socket_name_u = argv[6];
+	    if (strncmp("unix:", socket_name_u, 5) == 0)
+		socket_name_u += 5;
+	    if (argc == 8) {
+		notify_tag = argv[7];
+		len = url_unquote((uint8_t *)notify_tag, strlen(notify_tag));
+		if (len == -1) {
+		    rtpp_log_write(RTPP_LOG_ERR, cf->glog,
+		      "command syntax error - invalid URL encoding");
+		    reply_error(cf, controlfd, &raddr, rlen, cookie, 4);
+		    return 0;
+		}
+		notify_tag[len] = '\0';
+	    }
+	}
     }
     if (op == COPY) {
 	if (argc < 4 || argc > 5) {
@@ -565,6 +608,24 @@ handle_command(struct cfg *cf, int controlfd)
 		cp--;
 		break;
 
+	    case 'c':
+	    case 'C':
+		cp += 1;
+		for (t = cp; *cp != '\0'; cp++) {
+		    if (!isdigit(*cp) && *cp != ',')
+			break;
+		}
+		if (t == cp) {
+		    rtpp_log_write(RTPP_LOG_ERR, cf->glog, "command syntax error");
+		    reply_error(cf, controlfd, &raddr, rlen, cookie, 1);
+		    return 0;
+		}
+		codecs = alloca(cp - t + 1);
+		memcpy(codecs, t, cp - t);
+		codecs[cp - t] = '\0';
+		cp--;
+		break;
+
 	    default:
 		rtpp_log_write(RTPP_LOG_ERR, cf->glog, "unknown command modifier `%c'",
 		  *cp);
@@ -597,12 +658,24 @@ handle_command(struct cfg *cf, int controlfd)
 	}
     }
 
-    if (op != DELETE) {
+    /*
+     * Record and delete need special handling since they apply to all
+     * streams in the session.
+     */
+    switch (op) {
+    case DELETE:
+	i = handle_delete(cf, call_id, from_tag, to_tag, weak);
+	break;
+
+    case RECORD:
+	i = handle_record(cf, call_id, from_tag, to_tag);
+	break;
+
+    default:
 	i = find_stream(cf, call_id, from_tag, to_tag, &spa);
 	if (i != -1 && op != UPDATE)
 	    i = NOT(i);
-    } else {
-	i = handle_delete(cf, call_id, from_tag, to_tag, weak);
+	break;
     }
 
     if (i == -1 && op != UPDATE) {
@@ -622,6 +695,7 @@ handle_command(struct cfg *cf, int controlfd)
 
     switch (op) {
     case DELETE:
+    case RECORD:
 	reply_ok(cf, controlfd, &raddr, rlen, cookie);
 	return 0;
 
@@ -632,6 +706,13 @@ handle_command(struct cfg *cf, int controlfd)
 
     case PLAY:
 	handle_noplay(cf, spa, i);
+	if (strcmp(codecs, "session") == 0) {
+	    if (spa->codecs[i] == NULL) {
+		reply_error(cf, controlfd, &raddr, rlen, cookie, 6);
+		return 0;
+	    }
+	    codecs = spa->codecs[i];
+	}
 	if (playcount != 0 && handle_play(cf, spa, i, codecs, pname, playcount) != 0) {
 	    reply_error(cf, controlfd, &raddr, rlen, cookie, 6);
 	    return 0;
@@ -640,14 +721,22 @@ handle_command(struct cfg *cf, int controlfd)
 	return 0;
 
     case COPY:
-    case RECORD:
-	handle_record(cf, spa, i, recording_name);
+	handle_copy(cf, spa, i, recording_name);
 	reply_ok(cf, controlfd, &raddr, rlen, cookie);
 	return 0;
 
     case QUERY:
 	handle_query(cf, controlfd, &raddr, rlen, cookie, spa, i);
 	return 0;
+
+    case LOOKUP:
+    case UPDATE:
+	/* those are handled below */
+	break;
+
+    default:
+	/* Programmatic error, should not happen */
+	abort();
     }
 
     pidx = 1;
@@ -656,13 +745,11 @@ handle_command(struct cfg *cf, int controlfd)
 	assert(op == UPDATE || op == LOOKUP);
 	if (spa->fds[i] == -1) {
 	    j = ishostseq(cf->bindaddr[0], spa->laddr[i]) ? 0 : 1;
-	    if (create_listener(cf, spa->laddr[i], cf->nextport[j],
-	      &lport, fds) == -1) {
+	    if (create_listener(cf, spa->laddr[i], &lport, fds) == -1) {
 		rtpp_log_write(RTPP_LOG_ERR, spa->log, "can't create listener");
 		reply_error(cf, controlfd, &raddr, rlen, cookie, 7);
 		return 0;
 	    }
-	    cf->nextport[j] = lport + 2;
 	    assert(spa->fds[i] == -1);
 	    spa->fds[i] = fds[0];
 	    assert(spa->rtcp->fds[i] == -1);
@@ -680,7 +767,9 @@ handle_command(struct cfg *cf, int controlfd)
 	lport = spa->ports[i];
 	lia[0] = spa->laddr[i];
 	pidx = (i == 0) ? 1 : 0;
-	spa->ttl = cf->max_ttl;
+	spa->ttl_mode = cf->ttl_mode;
+	spa->ttl[0] = cf->max_ttl;
+	spa->ttl[1] = cf->max_ttl;
 	if (op == UPDATE) {
 	    rtpp_log_write(RTPP_LOG_INFO, spa->log,
 	      "adding %s flag to existing session, new=%d/%d/%d",
@@ -697,13 +786,11 @@ handle_command(struct cfg *cf, int controlfd)
 	  call_id, from_tag, weak ? "weak" : "strong");
 
 	j = ishostseq(cf->bindaddr[0], lia[0]) ? 0 : 1;
-	if (create_listener(cf, cf->bindaddr[j], cf->nextport[j], &lport,
-	  fds) == -1) {
+	if (create_listener(cf, cf->bindaddr[j], &lport, fds) == -1) {
 	    rtpp_log_write(RTPP_LOG_ERR, cf->glog, "can't create listener");
 	    reply_error(cf, controlfd, &raddr, rlen, cookie, 10);
 	    return 0;
 	}
-	cf->nextport[j] = lport + 2;
 
 	/*
 	 * Session creation. If creation is requested with weak flag,
@@ -724,8 +811,11 @@ handle_command(struct cfg *cf, int controlfd)
 	}
 	memset(spa, 0, sizeof(*spa));
 	memset(spb, 0, sizeof(*spb));
-	for (i = 0; i < 2; i++)
+	for (i = 0; i < 2; i++) {
 	    spa->fds[i] = spb->fds[i] = -1;
+	    spa->last_update[i] = 0;
+	    spb->last_update[i] = 0;
+	}
 	spa->call_id = strdup(call_id);
 	if (spa->call_id == NULL) {
 	    handle_nomem(cf, controlfd, &raddr, rlen, cookie, 13, ia,
@@ -757,9 +847,11 @@ handle_command(struct cfg *cf, int controlfd)
 	spb->fds[0] = fds[1];
 	spa->ports[0] = lport;
 	spb->ports[0] = lport + 1;
-	spa->ttl = cf->max_ttl;
-	spb->ttl = -1;
-	spa->log = rtpp_log_open("rtpproxy", spa->call_id, 0);
+	spa->ttl[0] = cf->max_ttl;
+	spa->ttl[1] = cf->max_ttl;
+	spb->ttl[0] = -1;
+	spb->ttl[1] = -1;
+	spa->log = rtpp_log_open(cf, "rtpproxy", spa->call_id, 0);
 	spb->log = spa->log;
 	spa->rtcp = spb;
 	spb->rtcp = NULL;
@@ -792,9 +884,39 @@ handle_command(struct cfg *cf, int controlfd)
 
 	rtpp_log_write(RTPP_LOG_INFO, spa->log, "new session on a port %d created, "
 	  "tag %s", lport, from_tag);
+	if (cf->record_all != 0) {
+	    handle_copy(cf, spa, 0, NULL);
+	    handle_copy(cf, spa, 1, NULL);
+	}
+    }
+
+    if (op == UPDATE) {
+	if (cf->timeout_handler.socket_name == NULL && socket_name_u != NULL)
+	    rtpp_log_write(RTPP_LOG_ERR, spa->log, "must permit notification socket with -n");
+	if (spa->timeout_data.notify_tag != NULL) {
+	    free(spa->timeout_data.notify_tag);
+	    spa->timeout_data.notify_tag = NULL;
+	}
+	if (cf->timeout_handler.socket_name != NULL && socket_name_u != NULL) {
+	    if (strcmp(cf->timeout_handler.socket_name, socket_name_u) != 0) {
+		rtpp_log_write(RTPP_LOG_ERR, spa->log, "invalid socket name %s", socket_name_u);
+		socket_name_u = NULL;
+	    } else {
+		rtpp_log_write(RTPP_LOG_INFO, spa->log, "setting timeout handler");
+		spa->timeout_data.handler = &cf->timeout_handler;
+		spa->timeout_data.notify_tag = strdup(notify_tag);
+	    }
+	} else if (socket_name_u == NULL && spa->timeout_data.handler != NULL) {
+	    spa->timeout_data.handler = NULL;
+	    rtpp_log_write(RTPP_LOG_INFO, spa->log, "disabling timeout handler");
+	}
     }
 
     if (ia[0] != NULL && ia[1] != NULL) {
+        if (spa->addr[pidx] != NULL)
+            spa->last_update[pidx] = dtime;
+        if (spa->rtcp->addr[pidx] != NULL)
+            spa->rtcp->last_update[pidx] = dtime;
 	/*
 	 * Unless the address provided by client historically
 	 * cannot be trusted and address is different from one
@@ -805,22 +927,42 @@ handle_command(struct cfg *cf, int controlfd)
 	  memcmp(ia[0], spa->addr[pidx], SA_LEN(ia[0])) == 0)) {
 	    rtpp_log_write(RTPP_LOG_INFO, spa->log, "pre-filling %s's address "
 	      "with %s:%s", (pidx == 0) ? "callee" : "caller", addr, port);
-	    if (spa->addr[pidx] != NULL)
-		free(spa->addr[pidx]);
+	    if (spa->addr[pidx] != NULL) {
+	        if (spa->canupdate[pidx] == 0) {
+	            if (spa->prev_addr[pidx] != NULL)
+	                 free(spa->prev_addr[pidx]);
+	            spa->prev_addr[pidx] = spa->addr[pidx];
+	        } else {
+		    free(spa->addr[pidx]);
+		}
+	    }
 	    spa->addr[pidx] = ia[0];
 	    ia[0] = NULL;
 	}
 	if (spa->rtcp->untrusted_addr[pidx] == 0 && !(spa->rtcp->addr[pidx] != NULL &&
 	  SA_LEN(ia[1]) == SA_LEN(spa->rtcp->addr[pidx]) &&
 	  memcmp(ia[1], spa->rtcp->addr[pidx], SA_LEN(ia[1])) == 0)) {
-	    if (spa->rtcp->addr[pidx] != NULL)
-		free(spa->rtcp->addr[pidx]);
+	    if (spa->rtcp->addr[pidx] != NULL) {
+	        if (spa->rtcp->canupdate[pidx] == 0) {
+	            if (spa->rtcp->prev_addr[pidx] != NULL)
+	                free(spa->rtcp->prev_addr[pidx]);
+	            spa->rtcp->prev_addr[pidx] = spa->rtcp->addr[pidx];
+	        } else {
+		    free(spa->rtcp->addr[pidx]);
+		}
+	    }
 	    spa->rtcp->addr[pidx] = ia[1];
 	    ia[1] = NULL;
 	}
     }
     spa->asymmetric[pidx] = spa->rtcp->asymmetric[pidx] = asymmetric;
     spa->canupdate[pidx] = spa->rtcp->canupdate[pidx] = NOT(asymmetric);
+    if (spa->codecs[pidx] != NULL) {
+	free(spa->codecs[pidx]);
+	spa->codecs[pidx] = NULL;
+    }
+    if (codecs != NULL)
+	spa->codecs[pidx] = strdup(codecs);
     if (requested_nsamples > 0) {
 	rtpp_log_write(RTPP_LOG_INFO, spa->log, "RTP packets from %s "
 	  "will be resized to %d milliseconds",
@@ -844,12 +986,13 @@ handle_command(struct cfg *cf, int controlfd)
 static int
 handle_delete(struct cfg *cf, char *call_id, char *from_tag, char *to_tag, int weak)
 {
-    int medianum, ndeleted;
+    int ndeleted;
+    unsigned int medianum;
     struct rtpp_session *spa, *spb;
     int cmpr, cmpr1, idx;
 
     ndeleted = 0;
-    for (spa = hash_table_findfirst(cf, call_id); spa != NULL;) {
+    for (spa = session_findfirst(cf, call_id); spa != NULL;) {
 	medianum = 0;
 	if ((cmpr1 = compare_session_tags(spa->tag, from_tag, &medianum)) != 0) {
 	    idx = 1;
@@ -859,7 +1002,7 @@ handle_delete(struct cfg *cf, char *call_id, char *from_tag, char *to_tag, int w
 	    idx = 0;
 	    cmpr = cmpr1;
 	} else {
-	    spa = hash_table_findnext(spa);
+	    spa = session_findnext(spa);
 	    continue;
 	}
 
@@ -881,15 +1024,15 @@ handle_delete(struct cfg *cf, char *call_id, char *from_tag, char *to_tag, int w
 	      spa->strong, spa->weak[0], spa->weak[1]);
 	    /* Skipping to next possible stream for this call */
 	    ++ndeleted;
-	    spa = hash_table_findnext(spa);
+	    spa = session_findnext(spa);
 	    continue;
 	}
 	rtpp_log_write(RTPP_LOG_INFO, spa->log,
-	  "forcefully deleting session %d on ports %d/%d",
+	  "forcefully deleting session %u on ports %d/%d",
 	   medianum, spa->ports[0], spa->ports[1]);
 	/* Search forward before we do removal */
 	spb = spa;
-	spa = hash_table_findnext(spa);
+	spa = session_findnext(spa);
 	remove_session(cf, spb);
 	++ndeleted;
 	if (cmpr != 2) {
@@ -947,7 +1090,7 @@ handle_play(struct cfg *cf, struct rtpp_session *spa, int idx, char *codecs,
 }
 
 static void
-handle_record(struct cfg *cf, struct rtpp_session *spa, int idx, char *rname)
+handle_copy(struct cfg *cf, struct rtpp_session *spa, int idx, char *rname)
 {
 
     if (spa->rrcs[idx] == NULL) {
@@ -962,6 +1105,29 @@ handle_record(struct cfg *cf, struct rtpp_session *spa, int idx, char *rname)
     }
 }
 
+static int
+handle_record(struct cfg *cf, char *call_id, char *from_tag, char *to_tag)
+{
+    int nrecorded, idx;
+    struct rtpp_session *spa;
+
+    nrecorded = 0;
+    for (spa = session_findfirst(cf, call_id); spa != NULL;
+      spa = session_findnext(spa)) {
+	if (compare_session_tags(spa->tag, from_tag, NULL) != 0) {
+	    idx = 1;
+	} else if (to_tag != NULL &&
+	  (compare_session_tags(spa->tag, to_tag, NULL)) != 0) {
+	    idx = 0;
+	} else {
+	    continue;
+	}
+	nrecorded++;
+	handle_copy(cf, spa, idx, NULL);
+    }
+    return (nrecorded == 0 ? -1 : 0);
+}
+
 static void
 handle_query(struct cfg *cf, int fd, struct sockaddr_storage *raddr,
   socklen_t rlen, char *cookie, struct rtpp_session *spa, int idx)
@@ -970,11 +1136,11 @@ handle_query(struct cfg *cf, int fd, struct sockaddr_storage *raddr,
     int len;
 
     if (cookie != NULL) {
-	len = sprintf(buf, "%s %d %lu %lu %lu %lu\n", cookie, spa->ttl,
+	len = sprintf(buf, "%s %d %lu %lu %lu %lu\n", cookie, get_ttl(spa),
 	  spa->pcount[idx], spa->pcount[NOT(idx)], spa->pcount[2],
 	  spa->pcount[3]);
     } else {
-	len = sprintf(buf, "%d %lu %lu %lu %lu\n", spa->ttl,
+	len = sprintf(buf, "%d %lu %lu %lu %lu\n", get_ttl(spa),
 	  spa->pcount[idx], spa->pcount[NOT(idx)], spa->pcount[2],
 	  spa->pcount[3]);
     }
