@@ -24,18 +24,20 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: rtpp_defines.h,v 1.9.2.1 2008/06/16 22:21:38 sobomax Exp $
+ * $Id: rtpp_defines.h,v 1.24 2009/01/12 10:55:27 sobomax Exp $
  *
  */
 
 #ifndef _RTPP_DEFINES_H_
 #define _RTPP_DEFINES_H_
 
+#include "config.h"
+
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/resource.h>
 #include <poll.h>
-#ifdef __linux__
+#ifdef HAVE_STDINT_H
 #include <stdint.h>
 #endif
 
@@ -54,8 +56,9 @@
 #define	TOS		0xb8
 #define	LBR_THRS	128	/* low-bitrate threshold */
 #define	CPORT		"22222"
-#define	POLL_LIMIT	100	/* maximum number of poll(2) calls per second */
+#define	POLL_LIMIT	200	/* maximum number of poll(2) calls per second */
 #define	LOG_LEVEL	RTPP_LOG_DBUG
+#define	UPDATE_WINDOW	10.0	/* in seconds */
 
 /* Dummy service, getaddrinfo needs it */
 #define	SERVICE		"34999"
@@ -63,7 +66,22 @@
 #define	CMD_SOCK	"/var/run/rtpproxy.sock"
 #define	PID_FILE	"/var/run/rtpproxy.pid"
 
-#define	rtpp_log_t	int
+/*
+ * TTL counters are used to detect the absence of audio packets
+ * in either direction.  When the counter reaches 0, the call timeout
+ * occurs.
+ */ 
+typedef enum {
+    TTL_UNIFIED = 0,		/* all TTL counters must reach 0 */
+    TTL_INDEPENDENT = 1		/* any TTL counter reaches 0 */
+} rtpp_ttl_mode;
+
+struct rtpp_timeout_handler {
+    char *socket_name;
+    int fd;
+    int connected;
+    char notify_buf[64];
+};
 
 struct cfg {
     int nodaemon;
@@ -72,7 +90,6 @@ struct cfg {
     int umode;			/* UDP control mode */
     int port_min;		/* Lowest UDP port for RTP */
     int port_max;		/* Highest UDP port number for RTP */
-    int nextport[2];
     struct rtpp_session **sessions;
     struct rtpp_session **rtp_servers;
     struct pollfd *pfds;
@@ -88,8 +105,12 @@ struct cfg {
      */
     struct sockaddr *bindaddr[2];	/* RTP socket(s) addresses */
     int tos;
+
     const char *rdir;
     const char *sdir;
+    int record_pcap;		/* Record in the PCAP format? */
+    int record_all;		/* Record everything */
+
     int rrtcp;			/* Whether or not to relay RTCP? */
     rtpp_log_t glog;
 
@@ -102,6 +123,19 @@ struct cfg {
     char *run_uname;
     char *run_gname;
     int no_check;
+
+    rtpp_ttl_mode ttl_mode;
+
+    struct rtpp_timeout_handler timeout_handler;
+
+    uid_t run_uid;
+    gid_t run_gid;
+
+    uint16_t port_table[65536];
+    int port_table_len;
+    int port_table_idx;
+
+    int log_level;
 };
 
 #endif
